﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace DuszaRegionalis.commands
{
    public class CommitCommand : Command
    {
        public string repoDir { get; set; }
        public string author { get; set; }
        public string commitMessage { get; set; }


        /// <summary>
        /// Akkor lehet rá szükség, ha az execute-ot nem szeretnénk futtatni.
        /// Jelen esetben ez akkor lehetséges, ha csak a dokumentáció érdekel.
        /// </summary>
        public CommitCommand()
        {

        }

        /// <summary>
        /// Létrehoz egy commit létrehozásáért felelős példányt, melynek függősége, hogy melyik repositorynak szeretnék megváltoztatni,
        /// ki szeretné azt megváltoztatni, valamint az új commit leírása
        /// </summary>
        /// <param name="repoDir">A repository elérési útja</param>
        /// <param name="author">A szerző neve</param>
        /// <param name="commitMessage">A commit leírása</param>
        public CommitCommand(string repoDir, string author, string commitMessage)
        {
            this.repoDir = repoDir;
            this.author = author;
            this.commitMessage = commitMessage;
        }
        /// <summary>
        /// Egy új commitot hoz létre, amennyiben lehetséges (inicializált a mappa és történt változás)
        /// A commitálás során legenerálódik a commit.details file
        /// </summary>
        void Command.execute()
        {
            if (!Program.initialized) throw new Exception("A parancs nem használható inicializálatlan mappák esetében.");
            string[] commitDetails = generateCommitDetails();
            if (commitDetails == null || commitDetails.Length == 0)
            {
                throw new Exception("Nem történt változás a fájlrendszeren.");
            }
            string newCommitId = (getLatestCommitNumber() + 1) + ".commit";
            string newCommitPath = repoDir + Program.VERSION_CONTROL_DIR + "\\" + newCommitId;
            copyDirectory(repoDir, newCommitPath);

            List<string> fileContent = new List<string>();
            fileContent.Add("Szulo: " + getParentId());
            fileContent.Add("Szerzo: " + author);
            fileContent.Add("Détum: " + DateTime.Now);
            fileContent.Add("Commit leírás: " + commitMessage);
            fileContent.Add("Változott:");
            fileContent.AddRange(commitDetails);
            Encoding utf8WithoutBom = new UTF8Encoding(false);
            File.WriteAllLines(newCommitPath + "\\commit.details", fileContent.ToArray(),utf8WithoutBom);
            File.WriteAllText(repoDir + Program.VERSION_CONTROL_DIR+"\\head.txt", newCommitId, utf8WithoutBom);

        }

        /// <summary>
        /// Legenerálja a változásokat a repository aktuális állapota és az aktív commit-ban lévő fileok összehasonlításával
        /// </summary>
        /// <returns>A változások tömbjével. Üres a tömb, amennyiben nincs változás.</returns>
        protected string[] generateCommitDetails()
        {
            List<string> commitDetails = new List<string>();
            int lastCommitNumber = getLatestCommitNumber();

            List<FileInfo> allFiles = getAllFiles(repoDir);
            List<string> allFilesPath = allFiles.Select(x => x.FullName).ToList();
            List<FileInfo> allFilesInParentCommit = new List<FileInfo>();
            List<string> allFilesPathInParentCommit = new List<string>();

            if (lastCommitNumber != 0)
            {
                allFilesInParentCommit = getAllFiles(repoDir + Program.VERSION_CONTROL_DIR + "\\" + lastCommitNumber + ".commit");
                allFilesInParentCommit.RemoveAll(x => x.Name == "commit.details");
            }

            allFilesInParentCommit.ForEach(parentFile => allFilesPathInParentCommit.Add(
                parentFile.FullName.Replace(Program.VERSION_CONTROL_DIR + "\\" + lastCommitNumber + ".commit", "")));

            for (int i=0; i< allFilesPath.Count;i++)
            {
                if (!allFilesPathInParentCommit.Contains(allFilesPath[i]))
                {
                    commitDetails.Add("új: " + allFilesPath[i] + " " + allFiles[i]);
                }
                else
                {
                    if(allFiles[i].LastWriteTime != allFilesInParentCommit[allFilesPathInParentCommit.IndexOf(allFilesPath[i])].LastWriteTime)
                    {
                        commitDetails.Add("változott: " + allFiles[i].FullName + " " + allFiles[i].LastWriteTime);
                    }
                }
            }

            for (int i=0; i< allFilesPathInParentCommit.Count;i++)
            {
                if (!allFilesPath.Contains(allFilesPathInParentCommit[i]))
                {
                    commitDetails.Add("törölt: " + allFilesInParentCommit[i].FullName + " " + allFilesInParentCommit[i].LastWriteTime);
                }
            }

            return commitDetails.ToArray();
        }

        /// <summary>
        /// kiolvassa a szülő azonosítóját a head.txt fájlból. Ha az nem létezik, feltételezi, hogy még nem volt commit, vagyis kötőjellel tér vissza
        /// </summary>
        /// <returns>A szülő commit azonosítójával</returns>
        protected string getParentId()
        {
            string headPath = repoDir + Program.VERSION_CONTROL_DIR + "\\head.txt";
            if (!File.Exists(headPath))
            {
                return "-";
            }
            Encoding ut8WithoutBom = new UTF8Encoding(false);
            string[] head = File.ReadAllLines(headPath, ut8WithoutBom);
            if (head.Length != 1)
            {
                throw new Exception("A head.txt több mint 1 sorból áll. Ez baj.");
            }
            return head[0];
        }
        /// <summary>
        /// Visszatér a megadott elérési út alatt tallálható fájlokkal. Kihagyja a head.txt-t és a commit-details, hogy ezekkel biztosan ne végezzünk akciókat indirekt.
        /// </summary>
        /// <param name="path">Egy mappa elérési útja, melynek filejaira kiváncsiak vagyunk.</param>
        /// <returns>A fileok listájával, kivéve a head.txt-t és a commit.details-t.</returns>
        protected List<FileInfo> getAllFiles(string path)
        {
            DirectoryInfo dirInfo = new DirectoryInfo(path);
            if (!dirInfo.Exists)
            {
                throw new Exception("A mappa nem létezik" + path);
            }
            List<FileInfo> files = dirInfo.GetFiles().ToList();
            files.Remove(new FileInfo("head.txt"));
            files.Remove(new FileInfo("commit.details"));

            return files;
        }
        /// <summary>
        /// A függvény ellenőrzi a meglévő commitok nevének helyességét a mappanevek alapján, majd visszatér az utolsó nevének . előtti részével
        /// </summary>
        /// <returns>Név szerint rendezve az utolsó mapp nevének első felével a verziókövetéses "rejtett" mappában.</returns>
        protected int getLatestCommitNumber()
        {
            DirectoryInfo dir = new DirectoryInfo(repoDir + Program.VERSION_CONTROL_DIR);
            DirectoryInfo[] subDirs = dir.GetDirectories();
            Array.Sort(subDirs, (d1, d2) => d1.Name.CompareTo(d2.Name));
            int commitNumber = 0;
            foreach (DirectoryInfo subDir in subDirs)
            {
                if (!subDir.Name.Contains(".")) throw new Exception("Egy mappa formátuma nem megfelelő. Nincs benne pont: " + subDir.Name);
                try
                {
                    commitNumber = Convert.ToInt32(subDir.Name.Split('.')[0]);
                }
                catch (FormatException)
                {
                    throw new Exception("Az egyik mappa nevében a pont előtt nem szám van: " + subDir.Name);
                }
            }
            return commitNumber;
        }
        /// <summary>
        /// Átmásol egy mappát egyik helyről a másikra. A másolás csak a fileokat veszi figyelembe (almappákat és azok elemeit nem)
        /// A másolás nem érinti a commit.details fájlt.
        /// </summary>
        /// <param name="source"> A forrás mappa, melynek elemei másolva lesznek</param>
        /// <param name="dest">A célhely, ahova másolni szeretnénk.</param>
        protected void copyDirectory (string source, string dest)
        {
            DirectoryInfo sourceDirInfo = new DirectoryInfo(source);

            if (!sourceDirInfo.Exists)
            {
                throw new Exception("A mappa nem létezik: " + source);
            }

            DirectoryInfo destDirInfo = new DirectoryInfo(source);

            if (!destDirInfo.Exists)
            {
                throw new Exception("A mappa nem létezik" + dest);
            }

            Directory.CreateDirectory(dest);
            FileInfo[] files = sourceDirInfo.GetFiles();
            foreach (FileInfo file in files)
            {
                if (file.Name != "commit.details")
                {
                    string tempPath = Path.Combine(dest, file.Name);
                    file.CopyTo(tempPath, false);
                }

            }
        }

        /// <summary>
        /// A commit parancs dokumentációját adja meg.
        /// </summary>
        /// <returns>A commit parancs dokumentációjával.</returns>
        string Command.getHelp()
        {
            return "Használat: commit\nEzzel a paranccsal tudsz létrehozni egy új commitot. A rendszer külön kérni fog egy nevet és egy commit message-et.";
        }
    }
}
