﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace DuszaRegionalis.commands
{
    /// <summary>
    /// Rekurzív adatszerkezet, hogy a commitokat képesek legyünk tárolni figyelembe véve, hogy ki volt a szülője
    /// Ehhez mérten egy commit szülője és gyerekei is commit fák.
    /// Megjegyzés: A szülők nem lenne feltétlen muszáj tárolni
    /// Megjegyzés 2: Jelen esetben 1 commithoz csak az azonosítóját tároljuk (a szülő és gyerek kivételével), ezt persze lehetne bővíteni, de nem éreztem szükségét,
    ///               Meg kell azonban jegyezni, hogy a legtisztább megoldás az lenne, ha minden beolvasásnál, kezelésnél már ezzel a struktőrával dolgoznánk
    /// </summary>
    class CommitTree
    {
        public string commitId { get; set; }
        public CommitTree parent { get; set; }

        public List<CommitTree> child { get; }

        /// <summary>
        /// Egy üres commit létrehozása, melyre csak az adatok inicializálása miatt van szükség.
        /// Egy "felhasználó" számára ennek nincs sok értelme, így privát a konstruktor. Mi a többi konstruktorban használjuk
        /// </summary>
        private CommitTree()
        {
            this.parent = null;
            this.child = new List<CommitTree>();
            this.commitId = "";
        }

        /// <summary>
        /// Egy olyan commitfa létrehozására való, melynél még nem tudjuk kik a gyerekek és ki a szülője
        /// Például Az első commit, melynek nincs szülője
        /// </summary>
        /// <param name="commitId">A commit azonosítója. A szülő és a gyerekek nem léteznek</param>
        public CommitTree(string commitId):this()
        {
            this.commitId = commitId;
        }

        /// <summary>
        /// Olyan commit lérehozására haszunáljuk, amikor már tudjuk a szülőt, de az még nem létezik a fában. Ez gyakorlatilag az első commitot jelenti
        /// </summary>
        /// <param name="parentId">A szülő azonosítója</param>
        /// <param name="commitId">A commit azonosítója</param>
        public CommitTree(string parentId, string commitId) : this(commitId)
        {
            CommitTree parent = new CommitTree(parentId);
            this.parent = parent;
            parent.addChild(this);
        }

        /// <summary>
        /// Egy olyan commit létrehozása, melynél már létezik faként a szülő
        /// </summary>
        /// <param name="parent">A szülő mint commitfa</param>
        /// <param name="commitId">A commit azonosítója</param>
        public CommitTree(CommitTree parent, string commitId) : this(commitId)
        {
            this.parent = parent;

        }

        /// <summary>
        /// Egy gyerek hozzáadása az adott commithoz.
        /// A láthatósága privát, mert ha valaki kívülről akarna hozzáadni gyereket, akkor eleve rosszul használná a struktúrát.
        /// Egy commit létrehozásakor ugyanis ha tudjuk a szülőt, akkor a konstruktor már eleve beilleszti a fába.
        /// </summary>
        /// <param name="commitTree">A commit mint fa, melyet hozzá akarunk adni a gyerekek listájához</param>
        private void addChild(CommitTree commitTree)
        {
            this.child.Add(commitTree);
        }


        /// <summary>
        /// Rekurzív metódus, mely megpróbálja az új commitot beilleszteni a fába, vagyis végigmegy a saját gyerekein és azok gyerekein és így tovább,
        /// amíg meg nem találja az új commit szülőjét, vagy el nem éri a fa alját.
        /// </summary>
        /// <param name="parentId">Az új commit szülőjének azonosítója</param>
        /// <param name="newElement">Az új commit azonosítója</param>
        /// <returns>Igaz, ha sikerült beilleszteni, hamis ha nem, vagyis ha nincs benne a fába a keresett szülő</returns>
        public bool putElementInTree(string parentId, string newElement)
        {
            if (this.commitId == parentId)
            {
                this.child.Add(new CommitTree(this,newElement));
                return true;
            }
            if (this.child.Count==0)
            {
                return false;
            }
            int i = 0;
            while (i < this.child.Count && ! this.child[0].putElementInTree(parentId, newElement)) { i++; }
            return i!=this.child.Count;
        }

        /// <summary>
        /// Legenrál egy commit erdőt. Az erdő fák listáját jelenti.
        /// A jelen esetben lista 1 elemű lesz, hiszen nem lehet a "közőjelre" több commitot építeni, mindennek az őse az 1. commit.
        /// A feladat szempontjából tehát ennek nincs értelme, csak érdekes feladatnak tartottam ezt megoldani. Valójában egyetlen commitfának
        /// reprezentálnia kell az összes commitot, különben valami nem stimmel.
        /// </summary>
        /// <param name="commits">Nyers commitazonosító-szülő párosokből csinál erdő struktúrát</param>
        /// <returns>A commit erdejével</returns>
        public static List<CommitTree> geenerateCommitForest(List<Commit> commits)
        {
            List<CommitTree> commitForest = new List<CommitTree>();
            foreach (Commit commit in commits)
            {
                int i = 0;
                while (i < commitForest.Count && !commitForest[i].putElementInTree(commit.parentId, commit.commitId))
                {
                    i++;
                }
                if (i == commitForest.Count)
                {
                    commitForest.Add(new CommitTree(commit.parentId, commit.commitId));
                }
            }
            return commitForest;
        }
    }
}
