﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace DuszaRegionalis.commands
{
    class ListRepositoryData : Command
    {
        public string options { get; set; }
        public string repoDir { get; set; }


        /// <summary>
        /// Akkor lehet rá szükség, ha az execute-ot nem szeretnénk futtatni.
        /// Jelen esetben ez akkor lehetséges, ha csak a dokumentáció érdekel.
        /// </summary>
        public ListRepositoryData()
        {

        }

        /// <summary>
        /// Létrehoz egy listázásért felelős példányt, melynek függősége, hogy melyik repositorynak szeretnék megváltoztatni,
        /// valamint hogy pontosan mit szeretnénk listázni.
        /// </summary>
        /// <param name="options">A listázási opció, mely lehet -all, -commitfa vagy egy commit azonosítója</param>
        /// <param name="repoDir">A repository elérési útja</param>
        public ListRepositoryData( string options, string repoDir)
        {
            this.options = options;
            this.repoDir = repoDir;
        }

        /// <summary>
        /// A megadott opciók alapján tud listázni/kiírni valamit közvetlenül a konzolra.
        /// </summary>
        void Command.execute()
        {
            if (!Program.initialized) throw new Exception("A parancs nem használható inicializálatlan mappák esetében.");
            switch (options)
            {
                case "-all":
                    getAllCommitList();
                    break;
                case "-commitfa":
                    printAllCommitsAsForest();
                    break;
                default:
                    printDetailsForGivenCommit();
                    break;
            }
        }
        private void getAllCommitList()
        {
            DirectoryInfo di = new DirectoryInfo(repoDir + Program.VERSION_CONTROL_DIR);

            foreach (DirectoryInfo dir in di.EnumerateDirectories())
            {
                Console.WriteLine(dir.Name);
            }
        }
        private void printDetailsForGivenCommit()
        {
            if (Directory.Exists(repoDir+ Program.VERSION_CONTROL_DIR + "\\" +options))
            {
                File.ReadAllLines(repoDir + Program.VERSION_CONTROL_DIR + "\\" + options + "\\commit.details").ToList().ForEach(line => Console.WriteLine(line));
            }
            else
            {
                throw new Exception("A commit nem létezik");
            }
        }

        /// <summary>
        /// Kirajzolja a commitok erdejét, ahol egy fa egy commitot és annak gyerekeit reprezentálja.
        /// Az elnevezés félrevezető, hiszen valójában nincsenek különálló commitok, mert az 1. commit kell hogy legyen minden commit őse,
        /// nem lehet kiválasztani a "kötőjel"-et.
        /// Megjegyzés: Az elágazások kirajzolását sokkal egyszerűbben is meg lehet oldani rekurzív adatszerkezetek, függvények nélkül is,
        ///             de be kell valjam elcsábított a feladat.
        /// </summary>
        private void printAllCommitsAsForest()
        {
            try
            {
                List<CommitTree> commitForest = CommitTree.geenerateCommitForest(readAllCommits());
                commitForest.ForEach(commitTree => printCommitAsTree(commitTree, "| "));
            }
            catch (Exception)
            {
                throw new Exception("Nem sikerült a commitok és azok szüleinek beolvasása. Talán nem jó a commit.details struktúrája.");
            }
        }

        /// <summary>
        /// Kirajzol egy commitot és gyerekeit úgy, hogy a gyereket a szülőhöz képest beljebb tolja. A függvény rekurzív
        /// </summary>
        /// <param name="commitTree">Egy commit mint fa adatszerkezet</param>
        /// <param name="prefix">Ami a commit neve elé szeretnék kiiratni. A beljebb tolás ennek változtatásával van megoldva, mikor a függvény meghívja saját magát</param>
        private void printCommitAsTree(CommitTree commitTree, string prefix)
        {
            Console.WriteLine(prefix + commitTree.commitId);
            if (commitTree.child != null)
            {
                commitTree.child.ForEach(child=>printCommitAsTree(child, "   " + prefix));
            }
        }

        /// <summary>
        /// Kiolvassa a commitok listáját a mappenevek és a commit.details-ben található szólőazonosító alapján
        /// </summary>
        /// <returns>A commit listájával, ahol egy commit tartalmazza a commit- valamint a szülőjánek azonosítóját</returns>
        List<Commit> readAllCommits()
        {
            List<Commit> commits = new List<Commit>();
            DirectoryInfo di = new DirectoryInfo(repoDir + Program.VERSION_CONTROL_DIR);

            foreach (DirectoryInfo dir in di.EnumerateDirectories())
            {
                string parentId = (File.ReadAllLines(dir.FullName + "\\commit.details"))[0].Split(' ')[1];
                commits.Add(new Commit(dir.Name, parentId));

            }
            return commits;
        }

        /// <summary>
        /// A list parancs dokumentációját adja meg.
        /// </summary>
        /// <returns>A list parancs dokumentációjával.</returns>
        string Command.getHelp()
        {
            return "Használat: list <opció>\nAz opciók a következők lehetnek:\n   " +
                "-all -> A commitok listája\n   " +
                "-commitfa -> Az elágazások megjelenítésére is szolgála erdő kirajzolása\n   " +
                "<commit azonosító> -> Egy commit azonosítója annak részleteinek megtekintéséhez";
        }
    }
}
